'use strict';

exports.__esModule = true;
exports.updateCurrValue = updateCurrValue;
exports.updateCurrVelocity = updateCurrVelocity;

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

var _react = require('react');

var _react2 = _interopRequireDefault(_react);

var _mapTree = require('./mapTree');

var _mapTree2 = _interopRequireDefault(_mapTree);

var _lodashIsplainobject = require('lodash.isplainobject');

var _lodashIsplainobject2 = _interopRequireDefault(_lodashIsplainobject);

var _stepper = require('./stepper');

var _stepper2 = _interopRequireDefault(_stepper);

var _noVelocity = require('./noVelocity');

var _noVelocity2 = _interopRequireDefault(_noVelocity);

var _mergeDiff = require('./mergeDiff');

var _mergeDiff2 = _interopRequireDefault(_mergeDiff);

var _animationLoop = require('./animationLoop');

var _animationLoop2 = _interopRequireDefault(_animationLoop);

var animationLoop = _animationLoop2['default']({
  // Fixed time step in seconds.
  timeStep: 1 / 60,
  // Slow-mo anyone? Give 0.1 a try.
  timeScale: 1,
  // Pause if we have more than this many steps worth of accumulated time.
  maxSteps: 10
});

function zero() {
  return 0;
}

// TODO: refactor common logic with updateCurrValue and updateCurrVelocity
function interpolateValue(alpha, nextValue, prevValue) {
  if (nextValue === null) {
    return null;
  }
  if (prevValue == null) {
    return nextValue;
  }
  if (typeof nextValue === 'number') {
    // https://github.com/chenglou/react-motion/pull/57#issuecomment-121924628
    return nextValue * alpha + prevValue * (1 - alpha);
  }
  if (nextValue.val != null && nextValue.config && nextValue.config.length === 0) {
    return nextValue;
  }
  if (nextValue.val != null) {
    var ret = {
      val: interpolateValue(alpha, nextValue.val, prevValue.val)
    };
    if (nextValue.config) {
      ret.config = nextValue.config;
    }
    return ret;
  }
  if (Array.isArray(nextValue)) {
    return nextValue.map(function (_, i) {
      return interpolateValue(alpha, nextValue[i], prevValue[i]);
    });
  }
  if (_lodashIsplainobject2['default'](nextValue)) {
    return Object.keys(nextValue).reduce(function (ret, key) {
      ret[key] = interpolateValue(alpha, nextValue[key], prevValue[key]);
      return ret;
    }, {});
  }
  return nextValue;
}

// TODO: refactor common logic with updateCurrVelocity

function updateCurrValue(frameRate, currValue, currVelocity, endValue, k, b) {
  if (endValue === null) {
    return null;
  }
  if (typeof endValue === 'number') {
    if (k == null || b == null) {
      return endValue;
    }
    // TODO: do something to stepper to make this not allocate (2 steppers?)
    return _stepper2['default'](frameRate, currValue, currVelocity, endValue, k, b)[0];
  }
  if (endValue.val != null && endValue.config && endValue.config.length === 0) {
    return endValue;
  }
  if (endValue.val != null) {
    var _ref = endValue.config || [170, 26];

    var _k = _ref[0];
    var _b = _ref[1];

    var ret = {
      val: updateCurrValue(frameRate, currValue.val, currVelocity.val, endValue.val, _k, _b)
    };
    if (endValue.config) {
      ret.config = endValue.config;
    }
    return ret;
  }
  if (Array.isArray(endValue)) {
    return endValue.map(function (_, i) {
      return updateCurrValue(frameRate, currValue[i], currVelocity[i], endValue[i], k, b);
    });
  }
  if (_lodashIsplainobject2['default'](endValue)) {
    return Object.keys(endValue).reduce(function (ret, key) {
      ret[key] = updateCurrValue(frameRate, currValue[key], currVelocity[key], endValue[key], k, b);
      return ret;
    }, {});
  }
  return endValue;
}

function updateCurrVelocity(frameRate, currValue, currVelocity, endValue, k, b) {
  if (endValue === null) {
    return null;
  }
  if (typeof endValue === 'number') {
    if (k == null || b == null) {
      return _mapTree2['default'](zero, currVelocity);
    }
    // TODO: do something to stepper to make this not allocate (2 steppers?)
    return _stepper2['default'](frameRate, currValue, currVelocity, endValue, k, b)[1];
  }
  if (endValue.val != null && endValue.config && endValue.config.length === 0) {
    return _mapTree2['default'](zero, currVelocity);
  }
  if (endValue.val != null) {
    var _ref2 = endValue.config || [170, 26];

    var _k = _ref2[0];
    var _b = _ref2[1];

    var ret = {
      val: updateCurrVelocity(frameRate, currValue.val, currVelocity.val, endValue.val, _k, _b)
    };
    if (endValue.config) {
      ret.config = endValue.config;
    }
    return ret;
  }
  if (Array.isArray(endValue)) {
    return endValue.map(function (_, i) {
      return updateCurrVelocity(frameRate, currValue[i], currVelocity[i], endValue[i], k, b);
    });
  }
  if (_lodashIsplainobject2['default'](endValue)) {
    return Object.keys(endValue).reduce(function (ret, key) {
      ret[key] = updateCurrVelocity(frameRate, currValue[key], currVelocity[key], endValue[key], k, b);
      return ret;
    }, {});
  }
  return _mapTree2['default'](zero, currVelocity);
}

var Spring = _react2['default'].createClass({
  displayName: 'Spring',

  propTypes: {
    endValue: _react.PropTypes.oneOfType([_react.PropTypes.func, _react.PropTypes.object, _react.PropTypes.array]).isRequired,
    children: _react.PropTypes.func.isRequired
  },

  getInitialState: function getInitialState() {
    var endValue = this.props.endValue;

    if (typeof endValue === 'function') {
      // TODO: provide warning for failing to provide base case
      endValue = endValue();
    }
    return {
      currValue: endValue,
      currVelocity: _mapTree2['default'](zero, endValue)
    };
  },

  componentDidMount: function componentDidMount() {
    this.startAnimating();
  },

  componentWillReceiveProps: function componentWillReceiveProps() {
    this.startAnimating();
  },

  unsubscribeAnimation: null,

  // used in animationRender
  hasUnmounted: false,

  componentWillUnmount: function componentWillUnmount() {
    if (this.unsubscribeAnimation) {
      this.unsubscribeAnimation();
      this.unsubscribeAnimation = null;
    }
    this.hasUnmounted = true;
  },

  startAnimating: function startAnimating() {
    if (!this.unsubscribeAnimation) {
      // means we're not animating
      this.unsubscribeAnimation = animationLoop.subscribe(this.animationStep, this.animationRender, this.state);
      animationLoop.start();
    }
  },

  animationStep: function animationStep(timeStep, state) {
    var currValue = state.currValue;
    var currVelocity = state.currVelocity;
    var endValue = this.props.endValue;

    if (typeof endValue === 'function') {
      endValue = endValue(currValue);
    }

    var newCurrValue = updateCurrValue(timeStep, currValue, currVelocity, endValue);
    var newCurrVelocity = updateCurrVelocity(timeStep, currValue, currVelocity, endValue);

    if (_noVelocity2['default'](currVelocity) && _noVelocity2['default'](newCurrVelocity)) {
      // check explanation in `animationRender`
      if (!this.hasUnmounted) {
        this.unsubscribeAnimation();
        this.unsubscribeAnimation = null;
      }
    }

    return {
      currValue: newCurrValue,
      currVelocity: newCurrVelocity
    };
  },

  animationRender: function animationRender(alpha, nextState, prevState) {
    // `this.hasUnmounted` might be true in the following condition:
    // user does some checks in `endValue` and calls an owner handler
    // owner sets state in the callback, triggering a re-render
    // re-render unmounts the Spring
    if (!this.hasUnmounted) {
      this.setState({
        currValue: interpolateValue(alpha, nextState.currValue, prevState.currValue),
        currVelocity: nextState.currVelocity
      });
    }
  },

  render: function render() {
    var currValue = this.state.currValue;

    return _react2['default'].Children.only(this.props.children(currValue));
  }
});

exports.Spring = Spring;
var TransitionSpring = _react2['default'].createClass({
  displayName: 'TransitionSpring',

  propTypes: {
    endValue: _react.PropTypes.oneOfType([_react.PropTypes.func, _react.PropTypes.objectOf({
      key: _react.PropTypes.any.isRequired
    })]).
    // coming soon
    // PropTypes.arrayOf(PropTypes.shape({
    //   key: PropTypes.any.isRequired,
    // })),
    // PropTypes.arrayOf(PropTypes.element),
    isRequired,
    willLeave: _react.PropTypes.oneOfType([_react.PropTypes.func]),

    // PropTypes.object,
    // PropTypes.array,
    // TODO: numbers? strings?
    willEnter: _react.PropTypes.oneOfType([_react.PropTypes.func]),

    // PropTypes.object,
    // PropTypes.array,
    children: _react.PropTypes.func.isRequired
  },

  getDefaultProps: function getDefaultProps() {
    return {
      willEnter: function willEnter(key, value) {
        return value;
      },
      willLeave: function willLeave() {
        return null;
      }
    };
  },

  getInitialState: function getInitialState() {
    var endValue = this.props.endValue;

    if (typeof endValue === 'function') {
      endValue = endValue();
    }
    return {
      currValue: endValue,
      currVelocity: _mapTree2['default'](zero, endValue)
    };
  },

  componentDidMount: function componentDidMount() {
    this.startAnimating();
  },

  componentWillReceiveProps: function componentWillReceiveProps() {
    this.startAnimating();
  },

  unsubscribeAnimation: null,

  // used in animationRender
  hasUnmounted: false,

  componentWillUnmount: function componentWillUnmount() {
    if (this.unsubscribeAnimation) {
      this.unsubscribeAnimation();
      this.unsubscribeAnimation = undefined;
    }
  },

  startAnimating: function startAnimating() {
    if (!this.unsubscribeAnimation) {
      this.unsubscribeAnimation = animationLoop.subscribe(this.animationStep, this.animationRender, this.state);
      animationLoop.start();
    }
  },

  animationStep: function animationStep(timeStep, state) {
    var currValue = state.currValue;
    var currVelocity = state.currVelocity;
    var endValue = this.props.endValue;
    var _props = this.props;
    var willEnter = _props.willEnter;
    var willLeave = _props.willLeave;

    if (typeof endValue === 'function') {
      endValue = endValue(currValue);
    }

    var mergedValue = undefined;
    // only other option is obj
    mergedValue = _mergeDiff2['default'](currValue, endValue,
    // TODO: stop allocating like crazy in this whole code path
    function (key) {
      return willLeave(key, currValue[key], endValue, currValue, currVelocity);
    });

    var hasNewKey = false;
    Object.keys(mergedValue).filter(function (key) {
      return !currValue.hasOwnProperty(key);
    }).forEach(function (key) {
      hasNewKey = true;
      var enterValue = willEnter(key, mergedValue[key], endValue, currValue, currVelocity);
      currValue[key] = enterValue;
      mergedValue[key] = enterValue;
      currVelocity[key] = _mapTree2['default'](zero, currValue[key]);
    });

    var newCurrValue = updateCurrValue(timeStep, currValue, currVelocity, mergedValue);
    var newCurrVelocity = updateCurrVelocity(timeStep, currValue, currVelocity, mergedValue);

    if (_noVelocity2['default'](currVelocity) && _noVelocity2['default'](newCurrVelocity) && !hasNewKey) {
      // check explanation in `Spring.animationRender`
      if (!this.hasUnmounted) {
        this.unsubscribeAnimation();
        this.unsubscribeAnimation = undefined;
      }
    }

    return {
      currValue: newCurrValue,
      currVelocity: newCurrVelocity
    };
  },

  animationRender: function animationRender(alpha, nextState, prevState) {
    // See comment in Spring.
    if (!this.hasUnmounted) {
      this.setState({
        currValue: interpolateValue(alpha, nextState.currValue, prevState.currValue),
        currVelocity: nextState.currVelocity
      });
    }
  },

  render: function render() {
    var currValue = this.state.currValue;

    return _react2['default'].Children.only(this.props.children(currValue));
  }
});
exports.TransitionSpring = TransitionSpring;