'use strict';

exports.__esModule = true;
exports['default'] = createAnimationLoop;

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

var _filter = require('./filter');

var _filter2 = _interopRequireDefault(_filter);

var _performanceNow = require('performance-now');

var _performanceNow2 = _interopRequireDefault(_performanceNow);

var _raf = require('raf');

var _raf2 = _interopRequireDefault(_raf);

function renderSubscriber(alpha, subscriber) {
  // subscriber.render: this.animationRender
  subscriber.render(alpha, subscriber.value, subscriber.prevValue);
  return subscriber.active;
}

var prototype = {
  running: false,
  shouldStop: false,
  lastTime: 0,
  accumulatedTime: 0,

  // step: this.animationStep
  // render: this.animationRender
  // value: state
  subscribe: function subscribe(step, render, value) {
    var subscriber = {
      value: value,
      prevValue: value,
      step: step,
      render: render,
      active: true
    };

    this.subscribers.push(subscriber);

    return function unsubscribe() {
      subscriber.active = false;
    };
  },

  loop: function loop() {
    var currentTime = _performanceNow2['default']();

    if (this.shouldStop) {
      this.running = this.shouldStop = false;
      return;
    }

    var timeStep = this.timeStep;
    // delta
    var frameTime = currentTime - this.lastTime;

    this.lastTime = currentTime;
    this.accumulatedTime += frameTime * this.timeScale;

    if (this.accumulatedTime > timeStep * this.maxSteps) {
      this.accumulatedTime = 0;
    }

    while (this.accumulatedTime > 0) {
      this.subscribers.forEach(this.step); // animationLoop.step
      this.accumulatedTime -= timeStep;
    }

    // Render and filter in one iteration.
    this.subscribers = _filter2['default'](this.subscribers, renderSubscriber, 1 + this.accumulatedTime / timeStep);

    if (this.subscribers.length === 0) {
      this.shouldStop = true;
    }

    _raf2['default'](this.loop);
  },

  start: function start() {
    if (this.subscribers.length) {
      if (this.shouldStop) {
        this.shouldStop = false;
      } else if (!this.running) {
        this.running = true;
        this.lastTime = _performanceNow2['default']();
        this.accumulatedTime = 0;
        _raf2['default'](this.loop);
      }
    }
  }

};

// stop() {
//   this.shouldStop = true;

//   return this;
// },

function createAnimationLoop(_ref) {
  var timeStep = _ref.timeStep;
  var timeScale = _ref.timeScale;
  var maxSteps = _ref.maxSteps;

  var animationLoop = Object.create(prototype);

  animationLoop.loop = animationLoop.loop.bind(animationLoop);
  animationLoop.subscribers = [];

  // timeStep is in milliseconds
  animationLoop.timeStep = timeStep * 1000; // seconds
  animationLoop.timeScale = timeScale;
  animationLoop.maxSteps = maxSteps;

  animationLoop.step = function (subscriber) {
    if (subscriber.active) {
      var value = subscriber.value; // value = this.state

      subscriber.prevValue = value;
      subscriber.value = subscriber.step(timeStep, value); // animationStep
    }
  };

  return animationLoop;
}

module.exports = exports['default'];